from io import IOBase

from . import ffi, librtmp
from .compat import byte_types
from .exceptions import RTMPError

__all__ = ["RTMPStream"]


class RTMPStream(IOBase):
    """A file-like interface to a stream within
       a RTMP session."""

    def __init__(self, client, update_buffer=True):
        self.client = client
        self._buf = self._view = None
        self._closed = False
        self._update_buffer = update_buffer
        self._updated_buffer = False

    def read(self, size):
        """Attempts to read data from the stream.

        :param size: int, The maximum amount of bytes to read.

        Raises :exc:`IOError` on error.
        """
        # If enabled tell the server that our buffer can fit the whole
        # stream, this often increases throughput alot.
        if self._update_buffer and not self._updated_buffer and self.duration:
            self.update_buffer((self.duration * 1000) + 5000)
            self._updated_buffer = True

        if not self._buf or len(self._buf) != size:
            self._buf = ffi.new("char[]", size)
            self._view = ffi.buffer(self._buf, size)

        res = librtmp.RTMP_Read(self.client.rtmp, self._buf, size)

        if res < 0:
            raise IOError("Failed to read data")

        return self._view[:res]

    def write(self, data):
        """Writes data to the stream.

        :param data: bytes, FLV data to write to the stream

        The data passed can contain multiple FLV tags, but it MUST
        always contain complete tags or undefined behaviour might
        occur.

        Raises :exc:`IOError` on error.
        """
        if isinstance(data, bytearray):
            data = bytes(data)

        if not isinstance(data, byte_types):
            raise ValueError("A bytes argument is required")

        res = librtmp.RTMP_Write(self.client.rtmp, data, len(data))

        if res < 0:
            raise IOError("Failed to write data")

        return res

    def pause(self):
        """Pauses the stream."""
        res = librtmp.RTMP_Pause(self.client.rtmp, 1)

        if res < 1:
            raise RTMPError("Failed to pause")

    def unpause(self):
        """Unpauses the stream."""
        res = librtmp.RTMP_Pause(self.client.rtmp, 0)

        if res < 1:
            raise RTMPError("Failed to unpause")

    def seek(self, time):
        """Attempts to seek in the stream.

        :param time: int, Time to seek to in seconds

        """
        res = librtmp.RTMP_SendSeek(self.client.rtmp, time)

        if res < 1:
            raise RTMPError("Failed to seek")

    def close(self):
        """Closes the connection."""
        if not self._closed:
            self._closed = True
            self.client.close()

    def update_buffer(self, ms):
        """Tells the server how big our buffer is (in milliseconds)."""
        librtmp.RTMP_SetBufferMS(self.client.rtmp, int(ms))
        librtmp.RTMP_UpdateBufferMS(self.client.rtmp)

    @property
    def duration(self):
        """The duration of the stream."""
        return librtmp.RTMP_GetDuration(self.client.rtmp)



from librtmp import RTMP, RTMPError

from pytest import raises


def test_connect():
    with raises(RTMPError):
        conn = RTMP("rtmp://localhost/app/playpath", live=True)
        conn.connect()


def test_remote_method():
    conn = RTMP("rtmp://localhost/app/playpath", live=True)
    my_remote_method = conn.remote_method("MyRemoteMethod", block=True)

    assert callable(my_remote_method) == True
    assert my_remote_method.__name__ == "MyRemoteMethod"



from cffi import FFI

__all__ = ["ffi"]

ffi = FFI()

preamble = """
    #include <librtmp/rtmp.h>
    #include <librtmp/log.h>

    /* The librtmp API is somewhat lacking. These functions
       help a bit. */

    void RTMP_SetSWFHash(RTMP *r, const char *swfhash,
                         uint32_t swfsize) {

        if (swfhash != NULL && swfsize > 0) {
            memcpy(r->Link.SWFHash, swfhash,
                   sizeof(r->Link.SWFHash));
            r->Link.SWFSize = swfsize;
        } else {
            r->Link.SWFSize = 0;
        }
    }

    int RTMP_GetInvokeCount(RTMP *r) {
        return r->m_numInvokes;
    }

    void RTMP_SetInvokeCount(RTMP *r, int count) {
        r->m_numInvokes = count;
    }

    /* Logging */
    #include <stdarg.h>

    void (*python_log_callback)(int level, char *msg);
    void c_log_callback(int level, const char *fmt, va_list args) {
        char buf[2048];
        vsprintf(buf, fmt, args);
        python_log_callback(level, buf);
    }

"""


ffi.cdef("""
    typedef struct AVal {
        char *av_val;
        int av_len;
    } AVal;

    typedef ... RTMP;
    typedef ... RTMPChunk;
    typedef int RTMP_LogLevel;
    typedef void* va_list;
    typedef void (RTMP_LogCallback)(int level, const char *fmt, va_list arg);

    typedef struct RTMPPacket {
        uint8_t m_headerType;
        uint8_t m_packetType;
        uint8_t m_hasAbsTimestamp;
        int m_nChannel;
        uint32_t m_nTimeStamp;
        int32_t m_nInfoField2;
        uint32_t m_nBodySize;
        uint32_t m_nBytesRead;
        RTMPChunk *m_chunk;
        char *m_body;
    } RTMPPacket;


    RTMP * RTMP_Alloc(void);
    void RTMP_Init(RTMP *r);
    int RTMP_SetupURL(RTMP *r, char *url);

    int RTMP_ParseURL(const char *url, int *protocol, AVal *host,
                      unsigned int *port, AVal *playpath, AVal *app);

    int RTMP_Connect(RTMP *r, RTMPPacket *cp);
    int RTMP_Serve(RTMP *r);

    int RTMP_IsConnected(RTMP *r);
    int RTMP_Socket(RTMP *r);
    int RTMP_IsTimedout(RTMP *r);
    int RTMP_ReadPacket(RTMP *r, RTMPPacket *packet);
    int RTMP_SendPacket(RTMP *r, RTMPPacket *packet, int queue);
    int RTMP_ClientPacket(RTMP *r, RTMPPacket *packet);

    int RTMP_ConnectStream(RTMP *r, int seek);
    int RTMP_ReconnectStream(RTMP *r, int seekTime);
    int RTMP_SendCreateStream(RTMP *r);
    void RTMP_DeleteStream(RTMP *r);
    int RTMP_Read(RTMP *r, char *buf, int size);
    int RTMP_Write(RTMP *r, const char *buf, int size);
    int RTMP_Pause(RTMP *r, int doPause);
    int RTMP_SendSeek(RTMP *r, int dTime);
    void RTMP_SetBufferMS(RTMP *r, int size);
    void RTMP_UpdateBufferMS(RTMP *r);

    int RTMP_LibVersion(void);
    int RTMP_SetOpt(RTMP *r, const AVal *opt, AVal *arg);
    double RTMP_GetDuration(RTMP *r);
    void RTMP_EnableWrite(RTMP *r);
    void RTMP_Close(RTMP *r);
    void RTMP_Free(RTMP *r);

    void RTMPPacket_Reset(RTMPPacket *p);
    void RTMPPacket_Dump(RTMPPacket *p);
    int RTMPPacket_Alloc(RTMPPacket *p, int nSize);
    void RTMPPacket_Free(RTMPPacket *p);


    RTMP_LogLevel RTMP_LogGetLevel(void);
    void RTMP_LogSetLevel(RTMP_LogLevel lvl);
    void RTMP_UserInterrupt(void);

    void RTMP_LogSetCallback(RTMP_LogCallback *cb);

    int RTMP_HashSWF(const char *url, unsigned int *size,
                     unsigned char *hash, int age);

    void RTMP_SetSWFHash(RTMP *r, const char *swfhash,
                         uint32_t swfsize);

    int RTMP_GetInvokeCount(RTMP *r);
    void RTMP_SetInvokeCount(RTMP *r, int count);

    typedef enum {
        AMF_NUMBER = 0, AMF_BOOLEAN, AMF_STRING, AMF_OBJECT,
        AMF_MOVIECLIP,      /* reserved, not used */
        AMF_NULL, AMF_UNDEFINED, AMF_REFERENCE, AMF_ECMA_ARRAY, AMF_OBJECT_END,
        AMF_STRICT_ARRAY, AMF_DATE, AMF_LONG_STRING, AMF_UNSUPPORTED,
        AMF_RECORDSET,      /* reserved, not used */
        AMF_XML_DOC, AMF_TYPED_OBJECT,
        AMF_AVMPLUS,        /* switch to AMF3 */
        AMF_INVALID = 0xff
    } AMFDataType;

    typedef ... AMFObjectProperty;
    typedef struct AMFObject { ...; } AMFObject;

    void AMF_Dump(AMFObject * obj);

    char *AMF_EncodeString(char *output, char *outend, const AVal * str);
    char *AMF_EncodeNumber(char *output, char *outend, double dVal);
    char *AMF_EncodeBoolean(char *output, char *outend, int bVal);
    char *AMF_EncodeInt16(char *output, char *outend, short nVal);
    char *AMF_EncodeInt24(char *output, char *outend, int nVal);
    char *AMF_EncodeInt32(char *output, char *outend, int nVal);

    int AMF_Decode(AMFObject * obj, const char *pBuffer, int nSize,
                   int bDecodeName);

    int AMF_CountProp(AMFObject * obj);
    AMFObjectProperty *AMF_GetProp(AMFObject * obj, const AVal * name,
                                   int nIndex);

    AMFDataType AMFProp_GetType(AMFObjectProperty * prop);
    void AMFProp_GetName(AMFObjectProperty * prop, AVal * name);
    void AMFProp_SetName(AMFObjectProperty * prop, AVal * name);
    double AMFProp_GetNumber(AMFObjectProperty * prop);
    int AMFProp_GetBoolean(AMFObjectProperty * prop);
    void AMFProp_GetString(AMFObjectProperty * prop, AVal * str);
    void AMFProp_GetObject(AMFObjectProperty * prop, AMFObject * obj);
    int AMFProp_IsValid(AMFObjectProperty * prop);

    void (*python_log_callback)(int level, char *msg);
    void *const c_log_callback;
""")


ffi.set_source("librtmp._librtmp", preamble,
               libraries=["rtmp"], sources=["src/librtmp/amf.c"])



from binascii import hexlify
from collections import namedtuple

from . import ffi, librtmp
from .aval import AVal
from .compat import bytes, str
from .exceptions import RTMPError

__all__ = ["add_signal_handler", "hash_swf"]


def add_signal_handler():
    """Adds a signal handler to handle KeyboardInterrupt."""
    import signal

    def handler(sig, frame):
        if sig == signal.SIGINT:
            librtmp.RTMP_UserInterrupt()
            raise KeyboardInterrupt

    signal.signal(signal.SIGINT, handler)


def hash_swf(url, age=30):
    hash = ffi.new("unsigned char[]", 32)
    size = ffi.new("unsigned int*")
    url = bytes(url, "utf8")

    res = librtmp.RTMP_HashSWF(url, size, hash, age)

    if res == 0:
        hash = hexlify(ffi.buffer(hash, 32)[:])
        size = size[0]

        return str(hash, "utf8"), size
    else:
        raise RTMPError("Failed to hash SWF")


RTMPURL = namedtuple("RTMPURL", ["protocol", "hostname",
                     "port", "playpath", "app"])


def parse_url(url):
    protocol = ffi.new("int*")
    hostname = AVal("")
    port = ffi.new("unsigned int*")
    playpath = AVal("")
    app = AVal("")

    res = librtmp.RTMP_ParseURL(bytes(url, "utf8"), protocol, hostname.aval, port,
                                playpath.aval, app.aval)

    if res < 1:
        result = RTMPURL(0, "", 0, "", "")
    else:
        result = RTMPURL(protocol[0], str(hostname.value, "utf8"), port[0],
                         str(playpath.value, "utf8"), str(app.value, "utf8"))

    return result



"""Python bindings for librtmp, built with cffi."""

from .compat import is_win32

if is_win32:
    import socket # Import socket to initialize WinSock

from ._librtmp import ffi, lib as librtmp

librtmp_version = librtmp.RTMP_LibVersion()

if librtmp_version < 0x020300:
    raise ImportError("Only librtmp version >= 2.3 is supported by this library")

from .exceptions import RTMPError, RTMPTimeoutError
from .logging import *
from .packet import *
from .rtmp import *
from .stream import *
from .utils import *

__title__ = "python-librtmp"
__version__ = "0.3.0"
__license__ = "Simplified BSD"
__author__ = "Christopher Rosell"
__copyright__ = "Copyright 2013-2015 Christopher Rosell"



try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from . import ffi, librtmp
from .aval import AVal
from .compat import is_py2, bytes, range
from .exceptions import AMFError

AMF_STRING_TYPES = (librtmp.AMF_STRING, librtmp.AMF_LONG_STRING)
AMF_OBJECT_DICT_TYPES = (librtmp.AMF_OBJECT, librtmp.AMF_ECMA_ARRAY)

__all__ = ["AMFObject", "decode_amf", "encode_amf"]


class AMFObject(dict):
    pass


def _create_buffer(size):
    pbuf = ffi.new("char[]", size)
    pend = pbuf + size
    buf = ffi.buffer(pbuf, size)

    return pbuf, pend, buf


def _encode_key_name(key):
    key = bytes(key, "utf8")
    key_len = len(key)

    pbuf, pend, buf = _create_buffer(key_len + 2)

    librtmp.AMF_EncodeInt16(pbuf, pend, key_len)
    buf[2:key_len + 2] = key

    return buf[:]


@singledispatch
def encoder(val):
    raise AMFError("Unable to encode '{0}' type".format(type(val).__name__))


@encoder.register(type(None))
def _encode_none(val):
    return bytes((librtmp.AMF_NULL,))


@encoder.register(str)
def _encode_str(val):
    val = AVal(val)
    pbuf, pend, buf = _create_buffer(val.aval.av_len + 1 + 4)

    res = librtmp.AMF_EncodeString(pbuf, pend, val.aval)
    size = res - pbuf

    return buf[:size]

if is_py2:
    encoder.register(unicode, _encode_str)


@encoder.register(float)
@encoder.register(int)
def _encode_number(val):
    val = float(val)
    pbuf, pend, buf = _create_buffer(9)

    librtmp.AMF_EncodeNumber(pbuf, pend, val)

    return buf[:]

if is_py2:
    encoder.register(long, _encode_number)


@encoder.register(bool)
def _encode_boolean(val):
    pbuf, pend, buf = _create_buffer(2)
    librtmp.AMF_EncodeBoolean(pbuf, pend, int(val))

    return buf[:]


@encoder.register(AMFObject)
def _encode_object(val):
    phead, headend, head = _create_buffer(4)
    head[0] = bytes((librtmp.AMF_OBJECT,))
    librtmp.AMF_EncodeInt24(phead + 1, headend, librtmp.AMF_OBJECT_END)

    body = bytearray()

    for key, value in val.items():
        body += _encode_key_name(key)
        body += encoder(value)

    return head[:1] + bytes(body) + head[1:]


@encoder.register(dict)
def _encode_ecma_array(val):
    phead, headend, head = _create_buffer(8)
    head[0] = bytes((librtmp.AMF_ECMA_ARRAY,))
    librtmp.AMF_EncodeInt32(phead + 1, headend, len(val))
    librtmp.AMF_EncodeInt24(phead + 5, headend, librtmp.AMF_OBJECT_END)

    body = bytearray()

    for key, value in val.items():
        body += _encode_key_name(key)
        body += encoder(value)

    return head[:5] + bytes(body) + head[5:]


@encoder.register(list)
def _encode_array(val):
    phead, headend, head = _create_buffer(5)
    head[0] = bytes((librtmp.AMF_STRICT_ARRAY,))
    librtmp.AMF_EncodeInt32(phead + 1, headend, len(val))

    body = bytearray()

    for value in val:
        body += encoder(value)

    return head[:] + bytes(body)


def _decode_prop(prop):
    prop_type = librtmp.AMFProp_GetType(prop)

    if prop_type == librtmp.AMF_NUMBER:
        val = librtmp.AMFProp_GetNumber(prop)
    elif prop_type in AMF_STRING_TYPES:
        aval = AVal()
        librtmp.AMFProp_GetString(prop, aval.aval)
        val = aval.value.decode("utf8", "ignore")
    elif prop_type == librtmp.AMF_BOOLEAN:
        val = bool(librtmp.AMFProp_GetBoolean(prop))
    elif prop_type in AMF_OBJECT_DICT_TYPES:
        if prop_type == librtmp.AMF_OBJECT:
            val = AMFObject()
        else:
            val = dict()

        for key, value in _decode_prop_obj(prop):
            val[key] = value
    elif prop_type == librtmp.AMF_STRICT_ARRAY:
        val = []
        for key, value in _decode_prop_obj(prop):
            val.append(value)
    else:
        val = None

    return val


def _decode_prop_obj(prop):
    obj = ffi.new("AMFObject*")
    librtmp.AMFProp_GetObject(prop, obj)

    prop_count = librtmp.AMF_CountProp(obj)

    for i in range(prop_count):
        prop = librtmp.AMF_GetProp(obj, ffi.NULL, i)

        key = AVal()
        librtmp.AMFProp_GetName(prop, key.aval)
        key = key.value.decode("utf8", "ignore")

        value = _decode_prop(prop)

        yield key, value


def encode_amf(value):
    return encoder(value)


def decode_amf(body):
    obj = ffi.new("AMFObject*")
    res = librtmp.AMF_Decode(obj, body, len(body), 0)

    if res == ffi.NULL:
        raise AMFError("Unable to decode AMF data")

    rval = []
    prop_count = librtmp.AMF_CountProp(obj)

    for i in range(prop_count):
        prop = librtmp.AMF_GetProp(obj, ffi.NULL, i)
        val = _decode_prop(prop)
        rval.append(val)

    return rval



from binascii import unhexlify
from time import time

from . import ffi, librtmp
from .aval import AVal
from .amf import encode_amf, decode_amf, AMFError
from .compat import bytes, string_types, integer_types
from .exceptions import RTMPError, RTMPTimeoutError
from .packet import RTMPPacket, PACKET_TYPE_INVOKE, PACKET_SIZE_MEDIUM
from .stream import RTMPStream
from .utils import hash_swf


__all__ = ["RTMP", "RTMPCall"]


class RTMP(object):
    """ A RTMP client session.

    :param url: str, A RTMP URL in the format `rtmp[t][e|s]://hostname[:port][/app[/playpath]]`.
    :param playpath: str, Overrides the playpath parsed from the RTMP URL.
    :param tcurl: str, URL of the target stream. Defaults to `rtmp[t][e|s]://host[:port]/app`.
    :param app: str, Name of application to connect to on the RTMP server.
    :param pageurl: str, URL of the web page in which the media was embedded.
    :param auth: str, Authentication string to be appended to the connect string.
    :param connect_data: This value will be encoded to AMF and added to the connect packet.
    :param swfhash: str, SHA256 hash of the decompressed SWF file (hexdigest).
    :param swfsize: int, Size of the decompressed SWF file.
    :param swfurl: str, URL of the SWF player for the media.
    :param swfvfy: bool, Calculate the correct swfhash and swfsize parameter
                   from the `swfurl` specified.
    :param flashver: str, Version of the Flash plugin used to run the SWF player.
    :param subscribe: str, Name of live stream to subscribe to. Defaults to `playpath`.
    :param token: str, Key for SecureToken response, used if the server requires
                  SecureToken authentication.
    :param live: bool, Specify that the media is a live stream.
    :param jtv: str, JSON token used by Twitch/Justin.tv servers.
    :param socks: str, Use the specified SOCKS4 proxy.
    :param start: int, Start at num seconds into the stream. Not valid for live streams.
    :param stop: int, Stop at num seconds into the stream.
    :param buffer: int, Set buffer time to num milliseconds. This is used to control
                   rate of data sent by FMS servers, not buffering of data. The default is 30000.
    :param timeout: int, Timeout the session after num seconds without receiving any data
                    from the server. The default is 30.
    """

    def __init__(self, url, playpath=None, tcurl=None, app=None, pageurl=None,
                 auth=None, swfhash=None, swfsize=None, swfurl=None, swfvfy=None,
                 flashver=None, subscribe=None, token=None, live=None, jtv=None,
                 connect_data=None, socks=None, start=None, stop=None, buffer=None,
                 timeout=None):
        def set_opt(key, val):
            if val is not None:
                self.set_option(key, val)

        self.rtmp = librtmp.RTMP_Alloc()

        if self.rtmp == ffi.NULL:
            raise MemoryError("Failed to allocate RTMP handle")

        librtmp.RTMP_Init(self.rtmp)

        self._options = dict()
        self._invoke_args = dict()
        self._invoke_handlers = dict()
        self._invoke_results = dict()
        self._connect_result = None

        self.url = None

        if swfurl and swfvfy:
            swfhash, swfsize = hash_swf(swfurl)

        if swfhash:
            digest = unhexlify(swfhash)
            librtmp.RTMP_SetSWFHash(self.rtmp, digest, swfsize)

        # Socks option must be set before setup_url.
        set_opt("socks", socks)

        self.setup_url(url)

        set_opt("playpath", playpath)
        set_opt("tcUrl", tcurl)
        set_opt("app", app)

        set_opt("swfUrl", swfurl)
        set_opt("pageUrl", pageurl)
        set_opt("auth", auth)
        set_opt("flashver", flashver)

        set_opt("subscribe", subscribe)
        set_opt("token", token)
        set_opt("jtv", jtv)
        set_opt("live", live)

        set_opt("start", start)
        set_opt("stop", stop)
        set_opt("buffer", buffer)
        set_opt("timeout", timeout)

        if isinstance(connect_data, (list, tuple)):
            for data in connect_data:
                self._parse_connect_data(data)
        elif connect_data is not None:
            self._parse_connect_data(connect_data)

    def _parse_connect_data(self, val):
        if isinstance(val, bool):
            self.set_option("conn", "B:{0}".format(int(val)))
        elif isinstance(val, string_types):
            self.set_option("conn", "S:{0}".format(val))
        elif isinstance(val, integer_types):
            self.set_option("conn", "N:{0}".format(val))
        elif isinstance(val, type(None)):
            self.set_option("conn", "Z:")
        elif isinstance(val, dict):
            self.set_option("conn", "O:1")
            for key, value in val.items():
                if isinstance(value, bool):
                    self.set_option("conn", "NB:{0}:{1}".format(key, int(value)))
                elif isinstance(value, string_types):
                    self.set_option("conn", "NS:{0}:{1}".format(key, value))
                elif isinstance(value, integer_types):
                    self.set_option("conn", "NN:{0}:{1}".format(key, value))
            self.set_option("conn", "O:0")

    def set_option(self, key, value):
        """Sets a option for this session.

        For a detailed list of available options see the librtmp(3) man page.

        :param key: str, A valid option key.
        :param value: A value, anything that can be converted to str is valid.

        Raises :exc:`ValueError` if a invalid option is specified.

        """

        akey = AVal(key)
        aval = AVal(value)

        res = librtmp.RTMP_SetOpt(self.rtmp, akey.aval, aval.aval)

        if res < 1:
            raise ValueError("Unable to set option {0}".format(key))

        self._options[akey] = aval

    def setup_url(self, url):
        r"""Attempt to parse a RTMP URL.

        Additional options may be specified by appending space-separated
        key=value pairs to the URL. Special characters in values may need
        to be escaped to prevent misinterpretation by the option parser.
        The escape encoding uses a backslash followed by two hexadecimal
        digits representing the ASCII value of the character. E.g., spaces
        must be escaped as `\\20` and backslashes must be escaped as `\\5c`.

        :param url: str, A RTMP URL in the format `rtmp[t][e|s]://hostname[:port][/app[/playpath]]`

        Raises :exc:`RTMPError` if URL parsing fails.

        """

        self.url = bytes(url, "utf8")

        res = librtmp.RTMP_SetupURL(self.rtmp, self.url)
        if res < 1:
            raise RTMPError("Unable to parse URL")

    def connect(self, packet=None):
        """Connect to the server.

        :param packet: RTMPPacket, this packet will be sent instead
                       of the regular "connect" packet.

        Raises :exc:`RTMPError` if the connect attempt fails.

        """

        if isinstance(packet, RTMPPacket):
            packet = packet.packet
        else:
            packet = ffi.NULL

        res = librtmp.RTMP_Connect(self.rtmp, packet)
        if res < 1:
            raise RTMPError("Failed to connect")

        return RTMPCall(self, 1.0)

    def create_stream(self, seek=None, writeable=False, update_buffer=True):
        """Prepares the session for streaming of audio/video
           and returns a :class:`RTMPStream` object.

        :param seek: int, Attempt to seek to this position.
        :param writeable: bool, Make the stream writeable instead of readable.
        :param update_buffer: bool, When enabled will attempt to speed up
                              download by telling the server our buffer can
                              fit the whole stream.

        Raises :exc:`RTMPError` if a stream could not be created.

        Usage::

          >>> stream = conn.create_stream()
          >>> data = stream.read(1024)
        """

        if writeable:
            librtmp.RTMP_EnableWrite(self.rtmp)

        # Calling handle_packet() on a connect result causes
        # librtmp to send a CreateStream call. This is not always
        # desired when using process_packets(), therefore we do it
        # here instead.
        if self._connect_result:
            self.handle_packet(self._connect_result)

        if not seek:
            seek = 0

        res = librtmp.RTMP_ConnectStream(self.rtmp, seek)
        if res < 1:
            raise RTMPError("Failed to start RTMP playback")

        return RTMPStream(self, update_buffer=update_buffer)

    @property
    def connected(self):
        """Returns True if connected to the server.

        Usage::

          >>> conn.connected
          True
        """

        return bool(librtmp.RTMP_IsConnected(self.rtmp))

    def read_packet(self):
        """Reads a RTMP packet from the server.

        Returns a :class:`RTMPPacket`.

        Raises :exc:`RTMPError` on error.
        Raises :exc:`RTMPTimeoutError` on timeout.

        Usage::

          >>> packet = conn.read_packet()
          >>> packet.body
          b'packet body ...'
        """

        packet = ffi.new("RTMPPacket*")
        packet_complete = False

        while not packet_complete:
            res = librtmp.RTMP_ReadPacket(self.rtmp, packet)

            if res < 1:
                if librtmp.RTMP_IsTimedout(self.rtmp):
                    raise RTMPTimeoutError("Timed out while reading packet")
                else:
                    raise RTMPError("Failed to read packet")

            packet_complete = packet.m_nBytesRead == packet.m_nBodySize

        return RTMPPacket._from_pointer(packet)

    def send_packet(self, packet, queue=True):
        """Sends a RTMP packet to the server.

        :param packet: RTMPPacket, the packet to send to the server.
        :param queue: bool, If True, queue up the packet in a internal queue rather
                      than sending it right away.

        """

        if not isinstance(packet, RTMPPacket):
            raise ValueError("A RTMPPacket argument is required")

        return librtmp.RTMP_SendPacket(self.rtmp, packet.packet,
                                       int(queue))

    def handle_packet(self, packet):
        """Lets librtmp look at a packet and send a response
           if needed."""

        if not isinstance(packet, RTMPPacket):
            raise ValueError("A RTMPPacket argument is required")

        return librtmp.RTMP_ClientPacket(self.rtmp, packet.packet)

    def process_packets(self, transaction_id=None, invoked_method=None,
                        timeout=None):
        """Wait for packets and process them as needed.

        :param transaction_id: int, Wait until the result of this
                               transaction ID is recieved.
        :param invoked_method: int, Wait until this method is invoked
                               by the server.
        :param timeout: int, The time to wait for a result from the server.
                             Note: This is the timeout used by this method only,
                             the connection timeout is still used when reading
                             packets.

        Raises :exc:`RTMPError` on error.
        Raises :exc:`RTMPTimeoutError` on timeout.

        Usage::

          >>> @conn.invoke_handler
          ... def add(x, y):
          ...   return x + y

          >>> @conn.process_packets()

        """

        start = time()

        while self.connected and transaction_id not in self._invoke_results:
            if timeout and (time() - start) >= timeout:
                raise RTMPTimeoutError("Timeout")

            packet = self.read_packet()

            if packet.type == PACKET_TYPE_INVOKE:
                try:
                    decoded = decode_amf(packet.body)
                except AMFError:
                    continue

                try:
                    method, transaction_id_, obj = decoded[:3]
                    args = decoded[3:]
                except ValueError:
                    continue

                if method == "_result":
                    if len(args) > 0:
                        result = args[0]
                    else:
                        result = None

                    self._invoke_results[transaction_id_] = result
                else:
                    handler = self._invoke_handlers.get(method)
                    if handler:
                        res = handler(*args)
                        if res is not None:
                            self.call("_result", res,
                                      transaction_id=transaction_id_)

                    if method == invoked_method:
                        self._invoke_args[invoked_method] = args
                        break

                if transaction_id_ == 1.0:
                    self._connect_result = packet
                else:
                    self.handle_packet(packet)
            else:
                self.handle_packet(packet)

        if transaction_id:
            result = self._invoke_results.pop(transaction_id, None)

            return result

        if invoked_method:
            args = self._invoke_args.pop(invoked_method, None)

            return args

    def call(self, method, *args, **params):
        """Calls a method on the server."""

        transaction_id = params.get("transaction_id")

        if not transaction_id:
            self.transaction_id += 1
            transaction_id = self.transaction_id

        obj = params.get("obj")

        args = [method, transaction_id, obj] + list(args)
        args_encoded = map(lambda x: encode_amf(x), args)
        body = b"".join(args_encoded)

        format = params.get("format", PACKET_SIZE_MEDIUM)
        channel = params.get("channel", 0x03)

        packet = RTMPPacket(type=PACKET_TYPE_INVOKE,
                            format=format, channel=channel,
                            body=body)

        self.send_packet(packet)

        return RTMPCall(self, transaction_id)

    def remote_method(self, method, block=False, **params):
        """Creates a Python function that will attempt to
           call a remote method when used.

        :param method: str, Method name on the server to call
        :param block: bool, Wheter to wait for result or not

        Usage::

          >>> send_usher_token = conn.remote_method("NetStream.Authenticate.UsherToken", block=True)
          >>> send_usher_token("some token")
          'Token Accepted'
        """

        def func(*args):
            call = self.call(method, *args, **params)

            if block:
                return call.result()

            return call

        func.__name__ = method

        return func

    def invoke_handler(self, func=None, name=None):
        if not callable(func):
            return lambda f: self.invoke_handler(func=f, name=func)

        method = name or func.__name__
        self.register_invoke_handler(method, func)

        return func

    def register_invoke_handler(self, method, func):
        self._invoke_handlers[method] = func

    def close(self):
        """Closes the connection to the server."""
        if self.connected:
            librtmp.RTMP_Close(self.rtmp)

    @property
    def transaction_id(self):
        return librtmp.RTMP_GetInvokeCount(self.rtmp)

    @transaction_id.setter
    def transaction_id(self, val):
        librtmp.RTMP_SetInvokeCount(self.rtmp, int(val))

    def __del__(self):
        librtmp.RTMP_Free(self.rtmp)


class RTMPCall(object):
    """A RTMP call.

    Contains the result of a :meth:`RTMP.call`.
    """

    def __init__(self, conn, transaction_id):
        self._result = None

        self.conn = conn
        self.done = False
        self.transaction_id = transaction_id

    def result(self, timeout=None):
        """Retrieves the result of the call.

        :param timeout: The time to wait for a result from the server.

        Raises :exc:`RTMPTimeoutError` on timeout.
        """
        if self.done:
            return self._result

        result = self.conn.process_packets(transaction_id=self.transaction_id,
                                           timeout=timeout)

        self._result = result
        self.done = True

        return result



from . import ffi, librtmp
from .utils import add_signal_handler

__all__ = ["set_log_level", "get_log_level",
           "set_log_output", "add_log_callback", "remove_log_callback",
           "LOG_CRIT", "LOG_ERROR", "LOG_WARNING",
           "LOG_INFO", "LOG_DEBUG", "LOG_DEBUG2",
           "LOG_ALL"]

(LOG_CRIT, LOG_ERROR, LOG_WARNING, LOG_INFO, LOG_DEBUG,
 LOG_DEBUG2, LOG_ALL) = range(1, 8)

_log_callbacks = set()
_log_level = LOG_ALL
_log_output = None


def set_log_level(level):
    """Sets log level."""

    global _log_level
    _log_level = level


def get_log_level():
    """Returns current log level."""
    return _log_level


def set_log_output(fd):
    """Sets log output to a open file-object."""
    global _log_output
    _log_output = fd


def add_log_callback(callback):
    """Adds a log callback."""
    global _log_callbacks

    if not callable(callback):
        raise ValueError("Callback must be callable")

    _log_callbacks.add(callback)
    return callback


def remove_log_callback(callback):
    """Removes a log callback."""
    global _log_callbacks
    _log_callbacks.remove(callback)


@ffi.callback("void(int,char*)")
def _log_callback(level, msg):
    msg = ffi.string(msg)
    msg = msg.decode("utf8", "ignore")

    for callback in _log_callbacks:
        callback(level, msg)

    if hasattr(_log_output, "write") and level <= _log_level:
        _log_output.write(msg + "\n")

librtmp.python_log_callback = _log_callback
librtmp.RTMP_LogSetCallback(librtmp.c_log_callback)

add_signal_handler()



from . import ffi, librtmp

__all__ = ["RTMPPacket",
           "PACKET_SIZE_LARGE", "PACKET_SIZE_MEDIUM",
           "PACKET_SIZE_SMALL", "PACKET_SIZE_MINIMUM",

           "PACKET_TYPE_CHUNK_SIZE", "PACKET_TYPE_BYTES_READ_REPORT",
           "PACKET_TYPE_CONTROL", "PACKET_TYPE_SERVER_BW",
           "PACKET_TYPE_CLIENT_BW", "PACKET_TYPE_AUDIO",
           "PACKET_TYPE_VIDEO", "PACKET_TYPE_FLEX_STREAM_SEND",
           "PACKET_TYPE_FLEX_SHARED_OBJECT", "PACKET_TYPE_FLEX_MESSAGE",
           "PACKET_TYPE_INFO", "PACKET_TYPE_SHARED_OBJECT",
           "PACKET_TYPE_INVOKE", "PACKET_TYPE_FLASH_VIDEO"]


(PACKET_SIZE_LARGE, PACKET_SIZE_MEDIUM, PACKET_SIZE_SMALL,
 PACKET_SIZE_MINIMUM) = range(4)

PACKET_TYPE_CHUNK_SIZE         = 0x01
PACKET_TYPE_BYTES_READ_REPORT  = 0x03
PACKET_TYPE_CONTROL            = 0x04
PACKET_TYPE_SERVER_BW          = 0x05
PACKET_TYPE_CLIENT_BW          = 0x06
PACKET_TYPE_AUDIO              = 0x08
PACKET_TYPE_VIDEO              = 0x09
PACKET_TYPE_FLEX_STREAM_SEND   = 0x0F
PACKET_TYPE_FLEX_SHARED_OBJECT = 0x10
PACKET_TYPE_FLEX_MESSAGE       = 0x11
PACKET_TYPE_INFO               = 0x12
PACKET_TYPE_SHARED_OBJECT      = 0x13
PACKET_TYPE_INVOKE             = 0x14
PACKET_TYPE_FLASH_VIDEO        = 0x16


class RTMPPacket(object):
    @classmethod
    def _from_pointer(cls, pointer):
        packet = cls.__new__(cls)
        packet.packet = pointer

        return packet

    def __init__(self, type, format, channel, timestamp=0,
                 absolute_timestamp=False, body=None):
        self.packet = ffi.new("RTMPPacket*")
        self.type = type
        self.format = format
        self.channel = channel
        self.timestamp = timestamp
        self.absolute_timestamp = absolute_timestamp

        if not body:
            body = b""

        self.body = body

    @property
    def format(self):
        """Format of the packet."""
        return self.packet.m_headerType

    @format.setter
    def format(self, value):
        self.packet.m_headerType = int(value)

    @property
    def type(self):
        """Type of the packet."""
        return self.packet.m_packetType

    @type.setter
    def type(self, value):
        self.packet.m_packetType = int(value)

    @property
    def channel(self):
        """Channel of the packet."""
        return self.packet.m_nChannel

    @channel.setter
    def channel(self, value):
        self.packet.m_nChannel = int(value)

    @property
    def timestamp(self):
        """Timestamp of the packet."""
        return self.packet.m_nTimeStamp

    @timestamp.setter
    def timestamp(self, value):
        self.packet.m_nTimeStamp = int(value)

    @property
    def absolute_timestamp(self):
        """True if the timestamp is absolute."""
        return bool(self.packet.m_hasAbsTimestamp)

    @absolute_timestamp.setter
    def absolute_timestamp(self, value):
        self.packet.m_hasAbsTimestamp = int(bool(value))

    @property
    def body(self):
        """The body of the packet."""
        view = ffi.buffer(self.packet.m_body, self.packet.m_nBodySize)

        return view[:]

    @body.setter
    def body(self, value):
        size = len(value)
        librtmp.RTMPPacket_Alloc(self.packet, size)

        view = ffi.buffer(self.packet.m_body, size)
        view[:] = value

        self.packet.m_nBodySize = size

    def dump(self):
        """Dumps packet to logger."""
        librtmp.RTMPPacket_Dump(self.packet)

    def __del__(self):
        librtmp.RTMPPacket_Free(self.packet)



#!/usr/bin/env python
from setuptools import setup
from sys import version_info


install_requires = ["cffi>=1.0.1"]

if not (version_info[0] == 3 and version_info[1] >= 4):
    install_requires.append("singledispatch")

readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")

setup(
    name="python-librtmp",
    version="0.3.0",
    description="Python bindings for librtmp, built with cffi",
    long_description=readme + "\n\n" + history,
    author="Christopher Rosell",
    author_email="chrippa@tanuki.se",
    url="https://github.com/chrippa/python-librtmp",
    license="Simplified BSD",
    packages=["librtmp"],
    cffi_modules=["librtmp/ffi.py:ffi"],
    install_requires=install_requires,
    setup_requires=["cffi>=1.0.1"],
    zip_safe=False,
    keywords="python-librtmp",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ]
)



from . import ffi
from .compat import bytes, integer_types, string_types


__all__ = ["AVal"]


class AVal(object):
    def __init__(self, value=None):
        self.aval = ffi.new("AVal *")

        if value is not None:
            self.value = value

    @property
    def value(self):
        buf = ffi.buffer(self.aval.av_val, self.aval.av_len)
        return buf[:]

    @value.setter
    def value(self, value):
        if isinstance(value, integer_types):
            value = str(value)
        if isinstance(value, string_types):
            value = bytes(value, "utf8")
        elif isinstance(value, bool):
            value = str(value).lower()

        self.value_str = ffi.new("char[]", value)
        self.aval.av_val = self.value_str
        self.aval.av_len = len(value)



import os
import sys

is_py2 = (sys.version_info[0] == 2)
is_py3 = (sys.version_info[0] == 3)
is_win32 = os.name == "nt"

if is_py2:
    _str = str
    str = unicode
    range = xrange

    byte_types = _str
    integer_types = (int, long)
    string_types = (_str, unicode)

    def bytes(b=None, enc="ascii"):
        if b is None:
            return ""
        elif isinstance(b, list) or isinstance(b, tuple):
            return "".join([chr(i) for i in b])
        else:
            return _str(b)

elif is_py3:
    bytes = bytes
    str = str
    range = range

    byte_types = bytes
    integer_types = int
    string_types = str

__all__ = ["is_py2", "is_py3", "is_win32", "str", "bytes", "range",
           "byte_types", "integer_types", "string_types"]
